@@ -2,7 +2,7 @@
 # @author: Matthäus G. Chajdas
 # @license: 2-clause BSD
 
-__version__ = '1.0.4'
+__version__ = '1.0.10'
 
 import io
 import os
@@ -15,14 +15,30 @@ class TemplateException(Exception):
 	def __init__(self, message, position):
 		Exception.__init__(self, message)
 		self._position = position
+		self.message = message
+
+	def __str__(self):
+		pos = self.GetPosition()
+		result = []
+		if pos.filename:
+			result.append (pos.filename)
+			result.append(':')
+		result.append(str(pos.line))
+		result.append(':')
+		result.append(str(pos.column))
+		result.append(':')
+
+		result.append(self.message)
+
+		return ''.join(result)
 
 	def GetPosition (self):
-		'''Get the position where the exception occured.
+		'''Get the position where the exception occurred.
 
 		:returns: An object with two fields, ``line`` and ``column``.'''
 		from collections import namedtuple
-		p = namedtuple ('position', ['line', 'column'])
-		return p (self._position [0], self._position [1])
+		p = namedtuple ('position', ['line', 'column', 'filename'])
+		return p (self._position [0], self._position [1], self._position [2])
 
 class InvalidFormatter(TemplateException):
 	'''An invalid formatter was encountered.
@@ -35,7 +51,11 @@ class InvalidToken(TemplateException):
 	pass
 
 class InvalidWhitespaceToken(TemplateException):
-	'''An invalid whitespace token was encountered.'''
+	'''Only for backwards compatibility. Will be removed in 2.x.'''
+	pass
+
+class InvalidNamedCharacterToken(InvalidWhitespaceToken):
+	'''An invalid named character token was encountered.'''
 	pass
 
 class InvalidBlock(TemplateException):
@@ -47,9 +67,10 @@ class TextStream:
 
 	The text stream is used for input only and keeps track of the current read
 	pointer position in terms of line/column numbers.'''
-	def __init__(self, text):
+	def __init__(self, text, *, filename=None):
 		self._text = text
 		self._length = len (text)
+		self._filename = filename
 		self.Reset ()
 
 	def Reset(self):
@@ -91,7 +112,7 @@ class TextStream:
 		'''Get the current read position as a pair (line, column).'''
 		assert self._line >= 1
 		assert self._column >= 1
-		return (self._line, self._column)
+		return self._line, self._column, self._filename
 
 	def Skip(self, length):
 		'''Skip a number of characters starting from the current position.'''
@@ -166,19 +187,38 @@ class Formatter:
 		assert self.IsBlockFormatter ()
 		pass
 
+_formatters = {}
+
+class _RegisterFormatter:
+	def __init__ (self, *args):
+		self.__aliases = args
+
+	def __call__ (self, cls):
+		for alias in self.__aliases:
+			if hasattr (cls, 'Create'):
+				_formatters [alias] = cls.Create
+			else:
+				_formatters [alias] = cls
+
+@_RegisterFormatter ('indent')
 class IndentFormatter(Formatter):
 	'''Indent a block using tabs.'''
+	@classmethod
+	def Create (cls, value):
+		return cls (int (value))
+
 	def __init__(self, depth):
 		Formatter.__init__(self, FormatterType.Block)
 		self._depth = depth
 		self._tabs = '\t' * depth
 
-	def OnBlockBegin(self, isFirst):
+	def OnBlockBegin(self, _):
 		return self._tabs
 
 	def Format(self, block):
 		return block.replace ('\n', '\n' + self._tabs)
 
+@_RegisterFormatter ('list-separator', 'separator', 'l-s')
 class ListSeparatorFormatter(Formatter):
 	'''Separate block entries.
 
@@ -193,11 +233,16 @@ class ListSeparatorFormatter(Formatter):
 		if not isLast:
 			return self._value
 
+@_RegisterFormatter ('width', 'w')
 class WidthFormatter(Formatter):
 	'''Align the value to a particular width.
 
 	Negative values align to the left (i.e., the padding is added on the left:
 	``'  42'``), positive values to the right (``'42  '``).'''
+	@classmethod
+	def Create (cls, value):
+		return cls (int (value))
+
 	def __init__(self, width):
 		Formatter.__init__(self, FormatterType.Value)
 		self._width = str(width) if width >= 0 else '>' + str(-width)
@@ -205,6 +250,7 @@ class WidthFormatter(Formatter):
 	def Format(self, text):
 		return str.format ("{0:" + self._width + "}", str(text))
 
+@_RegisterFormatter ('prefix')
 class PrefixFormatter(Formatter):
 	'''Add a prefix to a value.'''
 	def __init__(self, prefix):
@@ -214,6 +260,7 @@ class PrefixFormatter(Formatter):
 	def Format(self, text):
 		return self._prefix + str (text)
 
+@_RegisterFormatter ('suffix')
 class SuffixFormatter(Formatter):
 	'''Add a suffix to a value.'''
 	def __init__(self, suffix):
@@ -223,6 +270,7 @@ class SuffixFormatter(Formatter):
 	def Format(self, text):
 		return str (text) + self._suffix
 
+@_RegisterFormatter ('default')
 class DefaultFormatter(Formatter):
 	'''Emit the default if the value is None, otherwise the value itself.'''
 	def __init__(self, value):
@@ -235,6 +283,7 @@ class DefaultFormatter(Formatter):
 		else:
 			return text
 
+@_RegisterFormatter ('upper-case', 'uc')
 class UppercaseFormatter(Formatter):
 	'''Format a value as uppercase.'''
 	def __init__(self):
@@ -243,6 +292,7 @@ class UppercaseFormatter(Formatter):
 	def Format(self, text):
 		return str (text).upper ()
 
+@_RegisterFormatter ('escape-newlines')
 class EscapeNewlineFormatter(Formatter):
 	'''Escape embedded newlines.'''
 	def __init__(self):
@@ -251,6 +301,7 @@ class EscapeNewlineFormatter(Formatter):
 	def Format(self, text):
 		return str (text).replace ('\n', '\\n')
 
+@_RegisterFormatter ('escape-string')
 class EscapeStringFormatter(Formatter):
 	'''Escape embedded newlines, tabs and quotes.'''
 	def __init__(self):
@@ -262,6 +313,7 @@ class EscapeStringFormatter(Formatter):
 		text = text.replace ('\"', '\\"')
 		return text
 
+@_RegisterFormatter ('wrap-string')
 class WrapStringFormatter(Formatter):
 	'''Wrap strings with quotation marks.'''
 	def __init__(self):
@@ -273,6 +325,7 @@ class WrapStringFormatter(Formatter):
 		else:
 			return text
 
+@_RegisterFormatter ('cbool')
 class CBooleanFormatter(Formatter):
 	'''For booleans, write true or false to the output. Otherwise,
 	the input is just passed through.'''
@@ -285,6 +338,7 @@ class CBooleanFormatter(Formatter):
 		else:
 			return value
 
+@_RegisterFormatter ('hex')
 class HexFormatter(Formatter):
 	'''Write an integer as a hex literal (0x133F).'''
 	def __init__(self):
@@ -296,38 +350,20 @@ class HexFormatter(Formatter):
 		result = '0x' + result [2:].upper ()
 		return result
 
-def GetFormatter (name, value = None, position=None):
+def _GetFormatter (name, value = None, position=None):
 	'''Get a formatter.
 
 	If the formatter cannot be found, an exception is raised.'''
-	if name == 'width' or name == 'w':
-		return WidthFormatter (int (value))
-	elif name == 'prefix':
-		return PrefixFormatter (value)
-	elif name == 'suffix':
-		return SuffixFormatter (value)
-	elif name == 'list-separator' or name == 'l-s' or name == 'separator':
-		return ListSeparatorFormatter(value)
-	elif name == 'indent':
-		return IndentFormatter(int (value))
-	elif name == 'upper-case' or name == 'uc':
-		return UppercaseFormatter ()
-	elif name == 'default':
-		return DefaultFormatter (value)
-	elif name == 'wrap-string':
-		return WrapStringFormatter ()
-	elif name == 'cbool':
-		return CBooleanFormatter ()
-	elif name == 'escape-newlines':
-		return EscapeNewlineFormatter ()
-	elif name == 'escape-string':
-		return EscapeStringFormatter ()
-	elif name == 'hex':
-		return HexFormatter ()
-	else:
+	createFunction = _formatters.get (name, None)
+	if not createFunction:
 		raise InvalidFormatter("Invalid formatter '{0}'".format (name),
 			position)
 
+	if value:
+		return createFunction (value)
+	else:
+		return createFunction ()
+
 class Token:
 	'''Represents a single token.
 
@@ -338,9 +374,19 @@ class Token:
 		{{#Foo}} -> name = Foo, prefix = #
 		{{Bar:width=8}} -> name = Bar, prefix = None,
 							flags = {width:8}
+
+	The constructor checks if the formatter matches the token type. A block
+	formatter can be only applied to a block token, and a value formatter only
+	to a value.
 	'''
 
 	__validPrefixes = {'#', '/', '_', '>', '!'}
+	__namedCharacterTokens = {
+		'LEFT_BRACE': '{',
+		'RIGHT_BRACE': '}',
+		'NEWLINE': '\n',
+		'SPACE': ' '
+	}
 
 	def __init__(self, name, start, end, position):
 		for prefix in self.__validPrefixes:
@@ -366,13 +412,21 @@ class Token:
 			flags = tmp[separator+1:].split(':')
 
 			for flag in flags:
-				(key, value) = (None, None)
-				if (flag.find ('=') != -1):
+				if flag.find ('=') != -1:
 					(key, value) = flag.split('=')
 				else:
 					(key, value) = (flag, None)
 
-				self.__formatters.append (GetFormatter (key, value, position))
+				formatter = _GetFormatter (key, value, position)
+
+				if formatter.IsBlockFormatter () and not self.IsBlockStart ():
+					raise InvalidFormatter ("Requested block formatter '{}' on non-block. Only "
+						"block formatters can be used on blocks.".format (key), position)
+				elif formatter.IsValueFormatter () and not self.IsValue ():
+					raise InvalidFormatter ("Requested value formatter '{}' for non-value. Only "
+						"value formatters can be used with values.".format (key), position)
+
+				self.__formatters.append (formatter)
 
 	def GetName(self):
 		'''Get the name of this token.'''
@@ -406,11 +460,11 @@ class Token:
 		'''Return true if this token is a block-close token.'''
 		return self.__prefix == '/'
 
-	def IsWhiteSpaceToken (self):
-		'''Return true if this token is a whitespace directive.'''
+	def IsNamedCharacter (self):
+		'''Return true if this token is a named character token.'''
 		return self.__prefix == '_'
 
-	def IsIncludeToken (self):
+	def IsInclude (self):
 		'''Return true if this token is an include directive.'''
 		return self.__prefix == '>'
 
@@ -418,17 +472,16 @@ class Token:
 		'''Return true if this token is a self-reference.'''
 		return self.__name[0] == '.'
 
-	def EvaluateWhiteSpaceToken (self, position):
-		'''Get the content of this token if this token is a whitespace token.
+	def EvaluateNamedCharacterToken (self, position):
+		'''Get the content of this token if this token is an escape character token.
 
-		If the content is not a valid whitespace name, this function will raise
-		:py:class:`InvalidWhitespaceToken`.'''
-		if (self.__name == 'NEWLINE'):
-			return '\n'
-		elif (self.__name == 'SPACE'):
-			return ' '
+		If the content is not a valid character name, this function will raise
+		:py:class:`InvalidSpecialCharacterToken`.'''
+		result = self.__namedCharacterTokens.get(self.__name)
+		if result:
+			return result
 		else:
-			raise InvalidWhitespaceToken ("Unrecognized white-space token '{}'".format (self.__name),
+			raise InvalidNamedCharacterToken ("Unrecognized named character token '{}'".format (self.__name),
 				position)
 
 	def IsValue (self):
@@ -442,11 +495,11 @@ class IncludeHandler:
 class Template:
 	'''The main template class.'''
 
-	def __init__(self, template, includeHandler=None):
+	def __init__(self, template, includeHandler=None, *, filename=None):
 		'''includeHandler is used to support nested templates. It must support
 		a single function, Get(templateName) which returns a new Template
 		instance. See also TemplateRepository().'''
-		self.__input = TextStream (template)
+		self.__input = TextStream (template, filename=filename)
 		self.__includeHandler = includeHandler
 		self.__log = logging.getLogger('Lina.Template')
 
@@ -504,8 +557,7 @@ class Template:
 			return
 
 		for formatter in token.GetFormatters():
-			if formatter.IsValueFormatter():
-				value = formatter.Format (value)
+			value = formatter.Format (value)
 
 		if value is not None:
 			value = str (value)
@@ -540,6 +592,15 @@ class Template:
 			if blockName in items:
 				blockItems = items [blockName]
 
+				# Using name=None is an alternative way to skip a block
+				# A block specified using name=None is ignored, as if it was
+				# never defined
+				if blockItems is None:
+					if start.IsNegatedBlockStart ():
+						break
+					else:
+						return
+
 				if start.IsNegatedBlockStart ():
 					return
 				break
@@ -556,9 +617,9 @@ class Template:
 		# name=dict => name = [dict]
 		# name=None => name = [{}]
 		# otherwise, just copy
-		if (blockItems == None):
-			instanceCount = 1
+		if blockItems is None:
 			blockItems = [{}]
+			instanceCount = 1
 		elif isinstance(blockItems, collections.abc.Mapping):
 			blockItems = [blockItems]
 			instanceCount = 1
@@ -592,21 +653,20 @@ class Template:
 			isLast = ((i+1) == instanceCount)
 
 			if (i == 0):
-				current [blockName + "#First"] = None
+				current [blockName + "#First"] = {}
 			if ((i+1) < instanceCount):
-				current [blockName + "#Separator"] = None
+				current [blockName + "#Separator"] = {}
 			if ((i+1) == instanceCount):
-				current [blockName + "#Last"] = None
+				current [blockName + "#Last"] = {}
 
-			hasBlockFormatter = False
-			for formatter in [f for f in start.GetFormatters() if f.IsBlockFormatter()]:
-				hasBlockFormatter = True
+			formatters = start.GetFormatters ()
+			for formatter in formatters:
 				blockFormatterPrefix = formatter.OnBlockBegin (isFirst)
 				if blockFormatterPrefix:
 					outputStream.write (blockFormatterPrefix)
 
 			context.append (current)
-			if hasBlockFormatter:
+			if formatters:
 				# Write the string to a temporary stream if a block formatter is
 				# present
 				tmpStream = io.StringIO()
@@ -615,9 +675,8 @@ class Template:
 
 				tmpString = tmpStream.getvalue()
 
-				for formatter in start.GetFormatters():
-					if formatter.IsBlockFormatter():
-						tmpString = formatter.Format (tmpString)
+				for formatter in formatters:
+					tmpString = formatter.Format (tmpString)
 				outputStream.write (tmpString)
 			else:
 				# Directly render into output stream for maximum performance
@@ -625,11 +684,10 @@ class Template:
 
 			context.pop ()
 
-			if hasBlockFormatter:
-				for formatter in [f for f  in start.GetFormatters() if f.IsBlockFormatter()]:
-					blockFormatterSuffix = formatter.OnBlockEnd (isLast)
-					if blockFormatterSuffix:
-						outputStream.write (blockFormatterSuffix)
+			for formatter in formatters:
+				blockFormatterSuffix = formatter.OnBlockEnd (isLast)
+				if blockFormatterSuffix:
+					outputStream.write (blockFormatterSuffix)
 
 	def __ReadToken(self, inputStream):
 		'''Read a single token from the stream.'''
@@ -658,7 +716,7 @@ class Template:
 		'''Find the block end for a block named blockname in a stream, starting
 		from the current position.'''
 
-		nestedStack = []
+		nestedStack = [blockname]
 		while not inputStream.IsAtEnd ():
 			current = inputStream.Get()
 
@@ -670,17 +728,16 @@ class Template:
 					nestedStack.append (token.GetName ())
 
 				if token.IsBlockClose ():
-					if len(nestedStack) > 0:
-						if token.GetName () != nestedStack.pop ():
-							raise InvalidToken (
-								"Invalid block nesting for block '{}'".format (token.GetName ()),
-								inputStream.GetPosition ())
-					elif token.GetName () == blockname:
+					lastBlock = nestedStack.pop ()
+					if token.GetName () != lastBlock:
+						raise InvalidBlock (
+							"Cannot close block '{}' here. Last open block is '{}'".format (token.GetName (), lastBlock),
+							inputStream.GetPosition ())
+					
+					if not nestedStack:
 						return token
-					else:
-						break
 		raise InvalidBlock ("Could not find block end for '{}'".format (blockname),
-			inputStream.GetPosition  ())
+			inputStream.GetPosition ())
 
 	def __ExpandInclude (self, outputStream, token, itemStack):
 		'''Expand an include directive.'''
@@ -703,10 +760,13 @@ class Template:
 					blockEnd = self.__FindBlockEnd(inputStream, token.GetName ())
 					self.__ExpandBlock(inputStream, outputStream,
 									   token, blockEnd, itemStack)
-				elif token.IsWhiteSpaceToken():
-					outputStream.write (token.EvaluateWhiteSpaceToken(inputStream.GetPosition  ()))
-				elif token.IsIncludeToken ():
+				elif token.IsNamedCharacter():
+					outputStream.write (token.EvaluateNamedCharacterToken(inputStream.GetPosition  ()))
+				elif token.IsInclude ():
 					self.__ExpandInclude (outputStream, token, itemStack)
+				else:
+					raise InvalidToken ("Invalid token encountered",
+						inputStream.GetPosition ())
 			else:
 				# pass through to output
 				outputStream.write(current)
@@ -740,4 +800,4 @@ class TemplateRepository (IncludeHandler):
 
 	def Get(self, name):
 		content = open(os.path.join(self.dir, name + self.suffix)).read ()
-		return Template (content, self)
+		return Template (content, self, filename=name)
